### irredasm.s: IBM PC assembler routines for irred (below version 2.00)
###
### Copyright (C) 2000 Richard P. Brent.
###
### These routines come with absolutely no warranty.
### For conditions of use see the GNU General Public License
### and the comments in the calling program irred.c (version 3.10 or above).
###
### These routines do not do prefetching and are best on processors
### before P-III. For P-III and above, use irrednas.s
###
### Richard Brent (rpb@comlab.ox.ac.uk), last revised 20000716
###
### Assemble with	gcc -c irredasm.s

.file		"irredasm.s"

######################################## prefetch ###########################
###
### int prefetching();	
###
### Returns 1 if using nasm (and prefetching), 0 otherwise
###
.text
.balign		16
.global		prefetching
	.type   prefetching,@function
prefetching:
	xorl %eax, %eax		# return 0
	ret

################################## cpuinfo ##################################
###
### void cpuinfo(flag, results)
### int flag;
### int results[4];
###
### Returns the cpu information given by cpuid instruction in 
### (eax, ebx, edx, ecx) -> (results[0], .. , results[3])
### with input eax = flag.
###
.text
.balign		16
.global		cpuinfo
	.type   cpuinfo,@function
cpuinfo:	
	pusha
	movl  36(%esp), %eax	# flag -> eax
	cpuid			# see Intel documentation for effect
	movl 40(%esp), %edi
	movl %eax, 0(%edi)	# return results (eax, ebx, edx, ecx)
	movl %ebx, 4(%edi)	# in second argument array
	movl %edx, 8(%edi)
	movl %ecx, 12(%edi)
	popa
	ret

################################## checkmmx ##################################
###
### int checkmmx();
###
### Returns 1 if MMX instructions are usable, 0 if not
###
.text
.balign		16
.global		checkmmx
	.type   checkmmx,@function
checkmmx:	
	pushl %ebx
	pushl %ecx
	pushl %edx
	movl  $1, %eax
	cpuid
	movl %edx, %eax
	shrl $23, %eax		# check bit 23
	andl $1, %eax		# nonzero if MMX OK
	popl %edx
	popl %ecx
	popl %ebx
	ret

################################## reducer ##################################
#
# reducer.s, RPB, 20000707
#
# Equivalent to following C code (with the loop unrolled twice):
# except that works in the case rs zero
#
# void reducer(a, b, kt, rs, prev)	
# unsigned long *a, *b;
# int kt, rs;
# unsigned long *prev;		/* Previous -> last value of new */
#
#   {
#   int j;
#   unsigned long old, new;
#   for (new = *prev, j = kt; j >= 0; j--) {
#     old = new;
#     new = a[j];
#     b[j] ^= (new >> rs) | (old << (32-rs));
#     }
#   *prev = new;
#   return;  
#   }  
#
#
.text
.balign		16
.global reducer
	.type	 reducer,@function

### aptr:DWORD, bptr:DWORD, kt:DWORD, rs:DWORD, prevptr:DWORD
###  36          40          44        48        52       

# *a = %ebx, *b = %esi, 
# kt = %edi, rs = %ecx, temp = %ebp, 
# new = %edx (or %eax), old = %eax (or %edx)	

reducer:
	pusha
	movl 36(%esp),%ebx	 # *a
	movl 40(%esp),%esi	 # *b
	movl 44(%esp),%edi	 # j = kt
	movl 48(%esp),%ecx	 # rs (actually in cl)
	movl 52(%esp),%eax	 # *prev
	movl (%eax),%edx	 # new = *prev
	testl %edi,%edi		 # check for early exit
	jl redr_exit		 # exit if kt le 0
	test $1,%di		 # test low bit of kt
	jnz redr_init		 # jump to main loop if kt odd
				 # here j even, do one iteration
	movl %edx,%eax		 # old = new
	movl 0(%ebx,%edi,4),%edx # new = a[j]
	movl %edx,%ebp		 # temp = new
	shrd %cl,%eax,%ebp	 # double-word shift of (old,temp) right rs
	xorl %ebp,0(%esi,%edi,4) # b[j] ^= temp
	decl %edi		 # j--
	testl %edi,%edi		 # check for exit condition
	jl redr_exit		 # exit if j le 0

redr_init:
	movl 0(%ebx,%edi,4),%eax  # %eax = new = a[j] (old is %edx here)

	.align 4

redr_loop:			  # main loop, here j is odd
				  # %edx <-> %eax first half
	movl %eax,%ebp		  # temp = new
	shrd %cl,%edx,%ebp	  # double-word shift of (old,temp) right rs
	movl -4(%ebx,%edi,4),%edx # %edx = new = a[j] (now old is %eax)
	xorl %ebp,0(%esi,%edi,4)  # b[j] ^= temp
	movl %edx,%ebp		  # temp = new
	shrd %cl,%eax,%ebp	  # double-word shift of (old,temp) right rs
	movl -8(%ebx,%edi,4),%eax # preload %eax = new = next a[j]
	xorl %ebp,-4(%esi,%edi,4) # b[j] ^= temp
	addl $-2,%edi		  # j -= 2
	jnl redr_loop		  # loop if j non-negative

redr_exit:
	movl 52(%esp),%eax	 # *prev
	movl %edx,(%eax)	 # *prev = new
	popa
	ret

################################## squarem ##################################
###
### void squarem(unsigned long* a, int q1);
###
### Version of square for IBM PC, using MMX instructions.
### Written by Samuli Larvala (Samuli.Larvala@hut.fi) 20000629
###
### This is faster than square4 on recent PCs which implement MMX
### instructions efficiently
###
.data
.balign		16
C0:	.quad 0x0F0F0F0F0F0F0F0F
C1:	.quad 0x3333333333333333
C2:	.quad 0x5555555555555555
	
.text
.balign		16
.global		squarem
	.type	squarem,@function

squarem:	
### src:DWORD, len:DWORD
	pushl %ebp
	movl  %esp, %ebp
	pushl %ecx
	
	movl  12(%ebp), %ecx	# q1
	movl  8(%ebp), %eax	# *a
	testl %ecx, %ecx
	jz    sqr_exit

	incl  %ecx

### Load constants and square block
	movq -8(%eax,%ecx,4), %mm3	# Load value to square

### Test if length %ecx is odd (if so, do one 32-bit value)
	test $1, %cl
	movq C0, %mm0		# Load constants
	movq C1, %mm1
	movq C2, %mm2
	jz   sqr_mainloop
	
	movq %mm3, %mm4
	psrlq $4, %mm3
	
	punpckhbw %mm3, %mm4
	pand %mm0, %mm4		# 0x0H0H0H0H0H0H0H0H

	movq %mm4, %mm5
	psllq $2, %mm4
	
	por  %mm4, %mm5
	movq -12(%eax,%ecx,4), %mm3
	pand %mm1, %mm5
	
	movq %mm5, %mm4
	psllq $1, %mm5

	por  %mm5, %mm4
	pand %mm2, %mm4

	movq %mm4, -8(%eax,%ecx,8)
	decl %ecx
	jz   sqr_exit

	.align	4	

sqr_mainloop:

## mm0 = 0F0F0F0F0F0F0F0F
## mm1 = 3333333333333333
## mm2 = 5555555555555555
## mm3 = ABCDEFGHabcdefgh	

	movq %mm3, %mm4		## mm4 = ABCDEFGHabcdefgh
	psrlq $4, %mm3		## mm3 = 0ABCDEFGHabcdefg

	movq %mm4, %mm5		## mm5 = ABCDEFGHabcdefgh
	punpckhbw %mm3, %mm4	## mm4 = 0AABBCCDDEEFFGGH
	
	punpcklbw %mm3, %mm5	## mm5 = Haabbccddeeffggh
	pand %mm0, %mm4		## mm4 = 0A0B0C0D0E0F0G0H

	pand %mm0, %mm5		## mm5 = 0a0b0c0d0e0f0g0h

	movq -16(%eax,%ecx,4), %mm3	## Preload next 64-bits
	movq %mm4, %mm7

	movq %mm5, %mm6
	psllq $2, %mm4
	
	psllq $2, %mm5
	por  %mm4, %mm7
	
	por  %mm5, %mm6
	pand %mm1, %mm7

	pand %mm1, %mm6
	movq %mm7, %mm4
	
	movq %mm6, %mm5
	psllq $1, %mm7
	
	psllq $1, %mm6
	por  %mm7, %mm4
	
	por  %mm6, %mm5
	pand %mm2, %mm4
	
	movq %mm4, -8(%eax,%ecx,8)	# Store H
	pand %mm2, %mm5

	movq %mm5, -16(%eax,%ecx,8)	# Store L
	subl $2, %ecx
	jnz  sqr_mainloop

sqr_exit:
	popl %ecx
	popl %ebp
	emms				# Clear MMX state
	ret

################################## reducem ##################################
###
### typedef unsigned long ULONG; 
### ULONG reducem(ULONG* a, ULONG* b, ULONG* c, int r1, int r2, int kt);
###
### Version of reducea for IBM PC, using MMX instructions.
### Written by Samuli Larvala (Samuli.Larvala@hut.fi) 20000629
### Minor mods RPB 20000710
###
.text
.balign		16
.global		reducem
	.type	reducem,@function

reducem:	
### aptr:DWORD, bptr:DWORD, cptr:DWORD, r1:DWORD, r2:DWORD, kt:DWORD
###  36          40          44          48        52        56
	pusha
	movl 56(%esp), %edx	# kt
	pxor %mm2, %mm2		# old	(here in case early exit)
	movl 48(%esp), %ecx	# r1	(can be zero)
	testl %edx, %edx	# counter (j)
	jz   rdc_exit

	incl %edx
	movl 52(%esp), %eax	# r2	(can be zero)
	movl 36(%esp), %esi	# aptr
	movl 40(%esp), %ebx	# bptr
	movl 44(%esp), %edi	# cptr
	movd %ecx, %mm0		# mm0 = r1
	movd %eax, %mm1		# mm1 = r2
	xorl $63, %ecx
	xorl $63, %eax
	incl %ecx		# 64 - r1 = r1c
	incl %eax	       	# 64 - r2 = r2c

	test $1, %dl		# test if length %edx is odd
	jz   rdc_mainloop

	movd -4(%esi,%edx,4), %mm4	# a[j]
	movq -4(%ebx,%edx,4), %mm5	# b[j]
	movq -4(%edi,%edx,4), %mm6	# c[j]

	movq %mm4, %mm3
	psrlq %mm0, %mm4		# (new >> r1)
	movq %mm3, %mm2
	psrlq %mm1, %mm3		# (new >> r2)

	pxor %mm4, %mm5			# b[j] xor (new >> r1)
	pxor %mm3, %mm6			# c[j] xor (new >> r2)

	movq %mm5, -4(%ebx,%edx,4)	# store new b[j]
	movq %mm6, -4(%edi,%edx,4)	# store new c[j]
	decl %edx
	jz   rdc_exit

	.align 4

rdc_mainloop:
	movq -8(%esi,%edx,4), %mm3	# a[j]
	movd %ecx, %mm5			# r1c
	movq %mm2, %mm7			# old

	psllq %mm5, %mm2		# old << r1c
	movq %mm3, %mm6			# new
	movq -8(%ebx,%edx,4), %mm4	# b[j]

	psrlq %mm0, %mm3		# new >> r1
	movq -8(%edi,%edx,4), %mm5	# c[j]
	por  %mm2, %mm3			# (new>>r1) or (old<<r1c)

	movd %eax, %mm2			# r2c
	psllq %mm2, %mm7		# old << r2c
	pxor %mm3, %mm4			# b[j] xor ((new>>r1) or (old<<r1c))

	movq %mm6, %mm2			# new (new becomes old)
	psrlq %mm1, %mm6		# new >> r2

	movq %mm4, -8(%ebx,%edx,4)	# store b[j]
	por  %mm7, %mm6
	pxor %mm6, %mm5

	movq %mm5, -8(%edi,%edx,4)	# store c[j]
	addl $-2, %edx
	jnz  rdc_mainloop

rdc_exit:
	popa
	movd %mm2, %eax			# return last new
	emms				# clear MMX state
	ret

#############################################################################

### Fast modulo reduction of polynomial Q(x) (mod 2)
###
### Written 20000622 by Samuli Larvala (Samuli.Larvala@hut.fi)
###
### Modified by Richard Brent (rpb@comlab.ox.ac.uk) 20000629
### to use 32-bit memory XOR, which in some cases is faster.
###

.text
.balign		16
.global		reduce2
	.type	reduce2,@function

reduce2:	
### aptr:DWORD, bptr:DWORD, cptr:DWORD, r1:DWORD, r2:DWORD, kt:DWORD
###  36          40          44          48        52        56
###
### Assumes r-s > 63 (as for 64-bit version)
###
#
# Equivalent to following C code except unrolled twice (and assumes this
# does not cause any problems because of overlap of arrays a, b, c - hence
# the restriction r-s > 63). Also, it works in the case r1 or r2 zero.
#
# unsigned long reducea(a, b, c, r1, r2, kt)	
# unsigned long *a, *b, *c; int kt;
#  {int j; unsigned long old, new;
#  for (new = 0, j = kt; j GE 0; j--) {
#    old = new; new = a[j];
#    b[j] ^= (new>>r1) | (old<<(32-r1));
#    c[j] ^= (new>>r2) | (old<<(32-r2));
#    }
#  return(new);  
#  }  
#
	pusha
	movl 56(%esp), %edx	# kt
	pxor %mm2, %mm2		# old	(here in case early exit)
	movl 48(%esp), %ecx	# r1
	testl %edx, %edx	# counter (j)
	jz   rdc2_exit

	incl %edx
	movl 52(%esp), %eax	# r2
	movl 36(%esp), %esi	# aptr
	movl 40(%esp), %ebx	# bptr
	movl 44(%esp), %edi	# cptr
	movd %ecx, %mm0		# mm0 = r1
	movd %eax, %mm1		# mm1 = r2
	xorl $63, %ecx
	xorl $63, %eax
	incl %ecx		# r1c = 64 - r1
	incl %eax		# r2c = 64 - r2

	test $1, %dl		# test if length %edx is odd
	jz   rdc2_init

	movd -4(%esi,%edx,4), %mm4	# a[j]
	movq -4(%ebx,%edx,4), %mm5	# b[j]
	movq -4(%edi,%edx,4), %mm6	# c[j]

	movq  %mm4, %mm3
	psrlq %mm0, %mm4		# (new >> r1)
	movq  %mm3, %mm2
	psrlq %mm1, %mm3		# (new >> r2)

	pxor  %mm4, %mm5		# b[j] xor (new >> r1)
	pxor  %mm3, %mm6		# c[j] xor (new >> r2)

	movq  %mm5, -4(%ebx,%edx,4)	# store new b[j]
	movq  %mm6, -4(%edi,%edx,4)	# store new c[j]
	decl  %edx
	jz    rdc2_exit

rdc2_init:
	movd %ecx, %mm5			# r1c 
	movd %eax, %mm4			# r2c 

	.align 4

# See the nasm version for a different way of writing this loop
# (faster on the P-III)

rdc2_mainloop:	

	movq -8(%esi,%edx,4), %mm3	# new = (a[j+1], a[j])
	movq  %mm2, %mm7		# old
	psllq %mm5, %mm2		# old << r1c
	movq  %mm3, %mm6		# new
	psrlq %mm0, %mm3		# new >> r1
	por   %mm2, %mm3		# (new>>r1) | (old<<r1c)
	psllq %mm4, %mm7		# old << r2c
	movq  %mm6, %mm2		# new (new becomes old)
	psrlq %mm1, %mm6		# new >> r2
	por   %mm7, %mm6		# (new>>r2) | (old<<r2c)
	movd  %mm3, %ebp		# low 32 bits
	psrlq $32, %mm3			# shift right 32
	movd  %mm3, %ecx		# other 32 bits
	xorl  %ecx, -4(%ebx,%edx,4)	# b[j+1] ^= etc
	xorl  %ebp, -8(%ebx,%edx,4)	# b[j] ^= (new>>r1) | (old<<r1c)
	movd  %mm6, %eax		# low 32 bits
	psrlq $32, %mm6			# shift right 32
	movd  %mm6, %ecx		# other 32 bits
	xorl  %ecx, -4(%edi,%edx,4)	# c[j+1] ^= etc
	xorl  %eax, -8(%edi,%edx,4)	# c[j] ^= (new>>r2) | (old<<r2c)
	addl  $-2, %edx			# loop counter j -= 2
	jnz   rdc2_mainloop		# end inner loop

rdc2_exit:
        popa
	movd  %mm2, %eax		# return last new (32 bits)
	emms				# clear MMX registers
	ret

#############################################################################
